/*
 * Command line utility for making sense of C declarations.
 * Copyright © 2011-2012, 2020-2021 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <assert.h>

#include <getopt.h>
#include <gettext.h>
#include <readline.h>
#include <striconv.h>
#include <localcharset.h>
#include <mbswidth.h>

#include "cdecl99.h"
#include "cdecl.h"

static const char *progname = "cdecl99";

#include "options.h"
static const char sopts[] = SOPT_STRING;
static const struct option lopts[] = {
	LOPTS_INITIALIZER,
	{0}
};

static void print_version(void)
{
	const char *copysign = "(C)";
	void *convsign = NULL;

	if (ENABLE_NLS) {
		convsign = str_iconv("\xc2\xa9", "UTF-8", locale_charset());
		if (convsign)
			copysign = convsign;
	}

	puts(PACKAGE_STRING);
	printf("Copyright %s 2021 Nick Bowler.\n", copysign);
	puts("License GPLv3+: GNU GPL version 3 or any later version.");
	puts("This is free software: you are free to change and redistribute it.");
	puts("There is NO WARRANTY, to the extent permitted by law.");

	free(convsign);
}

static void print_usage(FILE *f)
{
	fprintf(f, _("Usage: %s [options]\n"), progname);
	if (f != stdout)
		fprintf(f, _("Try %s --help for more information.\n"),
		           progname);
}

static int
print_optstring(const struct option *opt, const struct lopt_help *help)
{
	char optstring[100];
	int w;

	if (!ENABLE_NLS)
		goto no_translate;

	if (opt->has_arg) {
		w = snprintf(optstring, sizeof optstring,
		             _("  -%c, --%s=%s"), opt->val, opt->name,
		             pgettext_expr(opt->name, help->arg));
	} else {
		w = snprintf(optstring, sizeof optstring,
		             _("  -%c, --%s"), opt->val, opt->name);
	}

	if (w < 0)
		goto no_translate;

	w = mbsnwidth(optstring, w, 0);
	printf("%s", optstring);
	goto out;

no_translate:
	if (opt->has_arg) {
		w = printf("  -%c, --%s=%s", opt->val, opt->name, help->arg);
	} else {
		w = printf("  -%c, --%s", opt->val, opt->name);
	}
out:
	if (w < 0 || w > 18) {
		putchar('\n');
		return 0;
	}

	return w;
}

/*
 * Print a string, with each line indented by i spaces.  The first line
 * will be indented by w fewer spaces (to account for the cursor being in
 * some other column).
 */
void print_block(const char *s, int i, int w)
{
	for (; *s; w = 0) {
		const char *nl = strchr(s, '\n');
		int n = (nl ? nl-s : -1);

		printf("%*s%.*s\n", i-w, "", n, s);
		if (!nl)
			break;

		s = nl+1;
	}
}

static void print_help(void)
{
	const struct option *opt;

	print_usage(stdout);

	puts(_("This is \"cdecl99\": a command-line tool for parsing and constructing\n"
               "complicated C declarations."));
	putchar('\n');

	puts(_("Options:"));
	for (opt = lopts; opt->name; opt++) {
		struct lopt_help help;
		int w;

		if (!lopt_get_help(opt, &help))
			continue;

		w = print_optstring(opt, &help);

		if (ENABLE_NLS)
			help.desc = pgettext_expr(opt->name, help.desc);

		print_block(help.desc, 20, w);
	}
	putchar('\n');

	puts(_("For more information, see the cdecl99(1) man page."));
	putchar('\n');

	/*
	 * TRANSLATORS: Please add *another line* indicating where users should
	 * report translation bugs.
	 */
	printf(_("Report bugs to <%s>.\n"), PACKAGE_BUGREPORT);
}

static bool is_blank_line(const char *line)
{
	for (size_t i = 0; line[i]; i++) {
		if (!isblank((unsigned char)line[i]))
			return false;
	}

	return true;
}

static int repl(void)
{
	char *line;

	for (; (line = readline("> ")); free(line)) {
		if (!is_blank_line(line))
			add_history(line);

		if (run_command(line, true) > 0)
			break;
	}

	free(line);
	return 0;
}

static int repl_cmdline(int argc, char **argv)
{
	int opt, rc, ret = 0;

	optind = 1;
	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		if (opt != 'e')
			continue;

		rc = run_command(optarg, false);
		if (rc < 0)
			ret = -1;
		else if (rc > 0)
			break;
	}

	return ret;
}

static int repl_noninteractive(void)
{
	int rc, ret = 0, saved_errno;
	char *line = NULL;
	size_t n;

	while (getline(&line, &n, stdin) >= 0) {
		char *c = strchr(line, '\n');
		if (c)
			*c = '\0';

		rc = run_command(line, false);
		if (rc < 0)
			ret = -1;
		else if (rc > 0)
			break;
	}

	saved_errno = errno;
	free(line);

	if (ferror(stdin)) {
		errno = saved_errno;
		perror("read error");
		return -1;
	}

	return ret;
}

/* Initialize gettext */
static void init_i18n(void)
{
	if (!ENABLE_NLS)
		return;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
}

int main(int argc, char **argv)
{
	bool show_intro = true, interactive = true, execute = false;
	const char *filename = NULL;
	int i, opt, rc;

	if (argc > 0)
		progname = argv[0];

	init_i18n();

	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch (opt) {
		case 'q':
			show_intro = false;
			break;
		case 'b':
			interactive = false;
			break;
		case 'i':
			interactive = true;
			break;
		case 'f':
			filename = optarg;
			break;
		case 'e':
			execute = true;
			break;
		case 'V':
			print_version();
			return EXIT_SUCCESS;
		case 'H':
			print_help();
			return EXIT_SUCCESS;
		default:
			print_usage(stderr);
			return EXIT_FAILURE;
		}
	}

	if (optind < argc) {
		fprintf(stderr, _("%s: excess command-line arguments:"),
		                progname);
		for (i = optind; i < argc; i++) {
			fprintf(stderr, " %s", argv[i]);
		}
		fprintf(stderr, "\n");
		print_usage(stderr);
		return EXIT_FAILURE;
	}

	/* --filename and --execute imply --batch. */
	if (filename || execute)
		interactive = false;

	/* --batch implies --quiet */
	if (interactive && show_intro)
		print_version();

	/* --execute supersedes --filename */
	if (filename && !execute) {
		if (!freopen(filename, "r", stdin)) {
			perror(filename);
			return EXIT_FAILURE;
		}
	}

	if (interactive)
		rc = repl();
	else if (execute)
		rc = repl_cmdline(argc, argv);
	else
		rc = repl_noninteractive();

	if (rc != 0)
		return EXIT_FAILURE;
	return EXIT_SUCCESS;
}
